/*
 *  diskdumpcmn.c
 *
 *  Copyright (C) 2005-2007  FUJITSU LIMITED
 *
 */
/* $Id: diskdumpcmn.c,v 1.34 2007/05/31 15:33:47 akira Exp $ */

#define _GNU_SOURCE
#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdbool.h>
#include <string.h>
#include <stdarg.h>
#include <time.h>
#include <errno.h>
#include <sys/stat.h>
#include <termio.h>
#include <sys/vfs.h>

#include "version.h"
#include "diskdumpcmn.h"

#define Stderr(x, ...)	fprintf(stderr, x, ## __VA_ARGS__)

static const u64 failed = (u64)-1;

const char *cmdname;

#define DEFAULT_WINSZ	80

struct option {
	char	*kind;	/* character used as an option */
	char	*cmnt;	/* description to an option */
};

static void set_opt(struct option *opt, char *kind, char *comment)
{
	opt->kind = kind;
	opt->cmnt = comment;
}

static void print_opt(struct option *opt, int n)
{
	int i;

	for (i = 0; i < n; i++) {
		if (opt[i].kind != NULL) {
			Stderr("\t%s:   %s\n", opt[i].kind, opt[i].cmnt);
		}
	}
}

void usage(void)
{
	int n = 0;
	const char *opts;
	struct option opt[] = {[0 ... 5] = {NULL, NULL}};
	int ret;
	char *tgt = "device";

	/*
	 * -v: some verbose outputs.
	 * -V: version information.
	 * Both of them are common options for each command.
	 */
	if (strcmp(cmdname, "diskdumpfmt") == 0) {
		opts = "[-v] [[-a] [-p] [-f]]|[[-c] [-i]]";
		set_opt(&opt[n++], "-a",
		"format all blocks of the dump partition");
		set_opt(&opt[n++], "-c",
		"check the status of the dump partition");
		set_opt(&opt[n++], "-f",
		"format the dump partition forcibly");
		set_opt(&opt[n++], "-i",
		"ignore that the diskdump module is not loaded, work with -c");
		set_opt(&opt[n++], "-p",
		"display progress for formatting");
		ret = 255;
	} else if (strcmp(cmdname, "savecore") == 0) {
		opts = "[-D] [-f] [-p] [-P pageflags] [-v] [-X]";
		set_opt(&opt[n++], "-D",
		"delete dump from the dump partition after saved");
		set_opt(&opt[n++], "-f",
		"create a dump file forcibly even if dump directory exists");
		set_opt(&opt[n++], "-p",
		"display progress for creating a dump file");
		set_opt(&opt[n++], "-P",
		"uncompress pages which have page flags specified by option");
		set_opt(&opt[n++], "-X",
		"delete dump from the dump partition without saving panic dump");
		ret = 1;
	} else if (strcmp(cmdname, "diskdumpctl") == 0) {
		opts = "[-u] [-v]";
		set_opt(&opt[n++], "-u", "unregister the device");
		ret = 1;
	} else if (strcmp(cmdname, "dumpfilter") == 0) {
		tgt = "oldvmcore newvmcore";
		opts = "[-f] [-l dump_level] [-p] [-v]";
		set_opt(&opt[n++], "-f",
				"overwrite an existing file");
		set_opt(&opt[n++], "-l",
				"specify a dump level (default: 19)");
		set_opt(&opt[n++], "-p",
				"display progress for creating a dump file");
		ret = 1;
	} else if (strcmp(cmdname, "diskdumpmsg") == 0) {
		tgt = "vmcore [messages]";
		opts = "[-f] [-p] [-v]";
		set_opt(&opt[n++], "-f",
				"output salvaged messages even if they are older than /var/log/messages");
		set_opt(&opt[n++], "-p",
				"output salvaged messages to stdout");
		ret = 1;
	} else {
		Stderr("Error: command '%s' does not match.\n", cmdname);
		exit(255);
	}

	Stderr("Usage: %s %s %s\n", cmdname, opts, tgt);
	Stderr("       %s -V\n", cmdname);

	Stderr("Available options:\n");
	print_opt(opt, n);	/* shown as "\toption:   description" */
	Stderr("\t-v:   provide some verbose outputs\n");
	Stderr("\t-V:   show %s version and exit\n", cmdname);

	exit(ret);
}

void version(void)
{
	Stderr("%s version %s\n", cmdname, CMD_VERSION);

	exit(0);
}

int convert_version_number(struct version_form *ver, const char *v_str)
{
	/* Initialize the version number as "0.0.0". */
	ver->major = 0;
	ver->minor = 0;
	ver->micro = 0;

	return sscanf(v_str, "%d.%d.%d", &ver->major, &ver->minor, &ver->micro);
}

/*
 * It gets some format parameters, sample_rate, block_order,
 * total_blocks and so on, from "/proc/diskdump".
 */
u64 scan_param_proc(const char *param_name)
{
	FILE *fp;
	char buf[BUFSIZ];
	u64 val;

	if ((fp = fopen(_PROC_DISKDUMP, "r")) == NULL) {
		return failed;
	}

	while (fgets(buf, BUFSIZ, fp) != NULL) {
		if (strstr(buf, param_name) != NULL) {
			if (sscanf(buf, "# %*s %Lu", &val) != 1) {
				fclose(fp);
				return failed;
			}
			fclose(fp);
			return val;	/* Got an exact value. */
		}
	}

	fclose(fp);
	return failed;
}

static inline int get_window_width(void)
{
	int rc;
	struct winsize winsz;

	rc = ioctl(0, TIOCGWINSZ, &winsz);
	if (rc < 0 || winsz.ws_col == 0) {
		return DEFAULT_WINSZ;
	} else {
		return winsz.ws_col;
	}
}

static inline void print_bar(int symbol, int len, int max)
{
	int i;

	if (!(0x21 <= symbol && symbol < 0x7f)) {
		return;
	}

	for (i = 0; i < len; i++) {
		Stderr("%c", symbol);
	}

	Stderr("%*c", max - len + 1, ' ');
}

static void print_progress(const char *item, int symbol, off_t now, off_t max)
{
	double progress;
	int bar_len, bar_avl;
	char *prefix = xasprintf("%s: ", item);
	char *suffix = NULL;

	progress = (double)now * 100.0
		 / (double)max;

	if (progress > 100.0) {
		progress = 100.0;
	}

	suffix = xasprintf("[%5.1f%%]", progress);

	bar_avl = get_window_width() - strlen(prefix) - strlen(suffix) - 1;
	bar_len = progress / 100 * bar_avl;

	Stderr(prefix);
	print_bar(symbol, bar_len, bar_avl);
	Stderr(suffix);
	Stderr("\r");
	free(prefix);
	free(suffix);
}

static inline void erase_bar(const char *item)
{
	int bar_avl;
	char *prefix = xasprintf("%s: ", item);
	char *suffix = xasprintf("[%5.1f%%]", 100.0);

	bar_avl = get_window_width() - strlen(prefix);

	Stderr(prefix);
	Stderr("%*c", bar_avl, ' ');
	Stderr("\r");
	Stderr(prefix);
	Stderr(suffix);
	free(prefix);
	free(suffix);
}

void progress_bar(const char *item, int symbol, off_t now, off_t max)
{
	time_t tm;
	static time_t last_time = 0;

	tm = CURRENT_DATE_TIME;

	if (tm - last_time < 1) {
		return;
	}

	last_time = tm;

	print_progress(item, symbol, now, max);
}

void progress_bar_post(const char *item, int symbol, off_t now, off_t max)
{
	print_progress(item, symbol, now, max);
	erase_bar(item);
	Stderr("\n");
}

int execute_script(const char *script, const char *arg)
{
	struct stat sbuf;
	char *filename = xstrdup(script);
	char *cmdline;
	int res = FAILED;

	if (stat(filename, &sbuf) == 0) {
		cmdline = xasprintf("%s '%s'", filename, arg ? : "");
		res = system(cmdline);
		free(cmdline);
	}

	free(filename);
	return res;
}

int rm_dir(const char *path)
{
	char *rm = xasprintf("/bin/rm -rf %s", path);
	int res;

	res = system(rm);

	free(rm);
	return res;
}

char *convert_ip_addr(uint32_t ip)
{
	return xasprintf("%d.%d.%d.%d",
			(ip >> 24) & 0xff,
			(ip >> 16) & 0xff,
			(ip >>  8) & 0xff,
			(ip	 ) & 0xff);
}

bool is_swapon(void)
{
	FILE *fp;
	char line[BUFSIZ];

	if ((fp = fopen("/proc/swaps", "r")) == NULL) {
		return false;
	}

	if (fgets(line, BUFSIZ, fp) == NULL) {
		fclose(fp);
		return false;
	}

	/* to handle kernel-2.4.x */
	if (start_with(line, "Filename")) { /* try again */
		if (fgets(line, BUFSIZ, fp) == NULL) {
			fclose(fp);
			return false;
		}
	}

	fclose(fp);
	return true;
}

int is_block_device(const char *device_name)
{
	struct stat stt;

	if (stat(device_name, &stt) < 0) {
		return FAILED;
	}

	return S_ISBLK(stt.st_mode);
}

static dev_t get_device_type(const char *device_name)
{
	struct stat stt;

	if (stat(device_name, &stt) < 0) {
		return (dev_t)-1;
	}

	return stt.st_rdev;
}

static int is_used(dev_t device_type, const char *filename)
{
	FILE *fp;
	char buf[BUFSIZ];
	char dev_name[BUFSIZ];
	dev_t dev_type = 0;
	enum {
		not_used,
		used
	};
	int ret = not_used;

	if ((fp = fopen(filename, "r")) == NULL) {
		return FAILED;
	}

	while (fgets(buf, BUFSIZ, fp) != NULL) {
		if (sscanf(buf, "%s %*s", dev_name) != 1) {
			continue;
		}
		if (!is_block_device(dev_name)) {
			continue;
		}
		if ((dev_type = get_device_type(dev_name)) == (dev_t)-1) {
			continue;
		}
		if (dev_type == device_type) {
			ret = used;
			break;
		}
	}

	fclose(fp);
	return ret;
}

int is_mounted(const char *device_name)
{
	dev_t device_type;
	int i;
	int ret;
	enum {
		ERROR = -1,
		NOT_MOUNTED,
		MOUNTED,
	};
	const char *file[] = {
		"/proc/mounts",	/* usually enough for kernel-2.6.x */
		"/etc/mtab"	/* able to get device that is on rootfs
				   on kernel-2.4.x */
	};

	if ((device_type = get_device_type(device_name)) == (dev_t)-1) {
		return ERROR;
	}

	for (i = 0; i < 2; i++) {
		ret = is_used(device_type, file[i]);
		if (ret == FAILED) {
			return ERROR;
		} else if (ret) {
			return MOUNTED;
		}
	}

	return NOT_MOUNTED;
}

char *get_date_time_str(const time_t t)
{
	struct tm *tm;
	char date[] = DATE_TIME_FORMAT, *date_time;

	tm = localtime(&t);
	strftime(date, sizeof(date), "%Y-%m-%d-%H:%M", tm);

	date_time = xstrdup(date);

	return date_time;
}

int diff_days(time_t tm1, time_t tm2)
{
	return difftime(tm1, tm2) / (60 * 60 * 24);
}

bool same_file(const char *src, const char *dst)
{
	struct stat	src_sts, dst_sts;
	struct statfs	src_stfs, dst_stfs;

	if (stat(src, &src_sts) < 0 || stat(dst, &dst_sts) < 0) {
		return false;
	}
	if (statfs(src, &src_stfs) < 0 || statfs(dst, &dst_stfs) < 0) {
		return false;
	}
	if (src_sts.st_ino != dst_sts.st_ino) {
		return false;
	}
	if (memcmp(&src_stfs.f_fsid, &dst_stfs.f_fsid, sizeof(src_stfs.f_fsid))
			!= 0) {
		return false;
	}

	return true;
}

bool does_exist(const char *file)
{
	struct stat st;

	return stat(file, &st) < 0 ? false : true;
}

bool dump_signature_is_good(const void *signature)
{
	return memcmp(signature, DISK_DUMP_SIGNATURE, DUMP_SIG_LEN) == 0;
}

bool part_signature_is_good(const void *signature)
{
	return memcmp(signature, DUMP_PARTITION_SIGNATURE, PART_SIG_LEN) == 0;
}

bool swap_signature_is_good(const void *signature)
{
	return memcmp(signature, SWAP_PARTITION_SIGNATURE, SWAP_SIG_LEN) == 0;
}

char *kernel_path(const char *kernel)
{
	char *path = NULL;

	if (start_with(kernel, "2.4")) {
		path = xasprintf("/boot/vmlinux-%s", kernel);
	} else if (start_with(kernel, "2.6")) {
		path = xasprintf("/usr/lib/debug/lib/modules/%s/vmlinux",
								kernel);
	}

	return path;
}

void die_func(const char *filename, int lineno, const char *fmt, ...)
{
	va_list ap;

	va_start(ap, fmt);
	Stderr("%s: ", cmdname);
	vfprintf(stderr, fmt, ap);
	Stderr("\n\tat line %d in %s\n", lineno, filename);

	exit(1);
}

void *read_alloc(int fd, size_t len)
{
	void *data = xmalloc(len);

	if (read(fd, data, len) < len) {
		free(data);
		return NULL;
	}

	return data;
}

int get_page_size(void)
{
	static int page_size;

	if (!page_size) {
		long ret = sysconf(_SC_PAGE_SIZE);

		if (ret < 0) {
			die("sysconf(_SC_PAGE_SIZE) failed");
		}
		page_size = (int)ret;
	}

	return page_size;
}

bool start_with(const char *str, const char *prefix)
{
	return strncmp(str, prefix, strlen(prefix)) == 0;
}

char *chop(char *line, const char *sep)
{
	static const char * const default_sep = "\n";
	char *p;

	if (sep == NULL) {
		sep = default_sep;
	}

	if ((p = strpbrk(line, sep)) != NULL) {
		*p = '\0';
	}

	return line;
}

bool is_later(const char *filename, time_t tm)
{
	struct stat st;

	if (stat(filename, &st) < 0) {
		die("stat(%s) failed with %s", filename, strerror(errno));
	}

	return st.st_mtime > tm;
}

bool is_super_user(void)
{
	return geteuid() == 0;
}

static dev_t get_device_id(struct stat sts)
{
	if (S_ISDIR(sts.st_mode) || S_ISLNK(sts.st_mode)) {
		return sts.st_dev;
	} else if (S_ISBLK(sts.st_mode) || S_ISCHR(sts.st_mode)) {
		return sts.st_rdev;
	} else {
		return (dev_t)-1;
	}
}

bool same_fs(const char *src, const char *dst)
{
	struct stat src_sts, dst_sts;
	dev_t src_dev_id, dst_dev_id;

	if (stat(src, &src_sts) < 0 || stat(dst, &dst_sts) < 0) {
		return false;
	}

	if ((src_dev_id = get_device_id(src_sts)) == (dev_t)-1) {
		return false;
	}
	if ((dst_dev_id = get_device_id(dst_sts)) == (dev_t)-1) {
		return false;
	}

	if (src_dev_id != dst_dev_id) {
		return false;
	}

	return true;
}
