/*
 *  diskdumpfmt.c
 *
 *  Copyright (C) 2004-2007  FUJITSU LIMITED
 *  Written by Akira Imamura (imamura.akira@jp.fujitsu.com)
 */
/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */
/* $Id: diskdumpfmt.c,v 1.109 2007/05/18 15:35:44 akira Exp $ */

#include <stdio.h>
#include <stdlib.h>
#include <stdbool.h>
#include <unistd.h>
#include "diskdump_format.h"

#define CMD	"diskdumpfmt"

static void check_mounted_device(const char *device_name)
{
	int ret = is_mounted(device_name);

	switch (ret) {
	case -1:
		Err("%s: unable to check on whether the device is mounted\n",
			device_name);
		exit(255);
	case 0:
		/*
		 * the device is not mounted anywhere, so
		 * it could be checked or formatted.
		 */
		return;
	case 1:
		Err("%s: mounted device\n\
		(unable to do anything whenever the device is mounted)\n",
			device_name);
		exit(MOUNTED);
	default:
		Err("returned an unexpected value '%d'.\n", ret);
		exit(255);
	}
}

static void init_format_parameters(dump_device_t *device)
{
	int num;	/* the number of parameter values gotten */

	/*
	 * 3 values (sample rate, block order, and total blocks)
	 * are always required for quick formatting.
	 */
	num = get_module_params(device);
	switch (num) {
	case 0:
		Err("diskdump module may not have been loaded.\n\
		(only full formatting is available)\n");
		return;
	case 1:
		Err("unable to get module parameter \"block order\".\n");
		exit(255);
	case 2:
		Err("unable to get module parameter \"total blocks\".\n");
		exit(255);
	case 3:
		/* got all required parameter values. */
		return; /* both full and quick formatting available. */
	default:
		Err("returned an unexpected value '%d'.\n", num);
		exit(255);
	}
}

int main(int argc, char *argv[])
{
	dump_device_t *device;
	const char *device_name;
	int opt;
	bool chk_flg = false, frc_flg = false, all_flg = false, prg_flg = false;
	bool verbose = false, ignore = false;
	off_t fmtd_blk;
	extern int symbol;	/* for progress bar */

	cmdname = CMD;

	while ((opt = getopt(argc, argv, "acfipvV")) != -1) {
		switch (opt) {
		case 'a':
			all_flg = true;
			break;
		case 'c':
			chk_flg = true;
			break;
		case 'f':
			frc_flg = true;
			break;
		case 'i':
			ignore = true;
			break;
		case 'p':
			prg_flg = true;
			symbol = '#';
			break;
		case 'v':
			verbose = true;
			break;
		case 'V':
			version();
			break;
		default:
			usage();
			break;
		}
	}

	if ((ignore && !chk_flg)
	|| ((chk_flg || ignore) && (frc_flg || prg_flg || all_flg))) {
		usage();
	}

	if (optind + 1 != argc) {
		usage();
	}

	device_name = argv[optind];

	switch (is_block_device(device_name)) {
	case FAILED:
		Err("unable to open %s\n", device_name);
		exit(255);
	case 0:
		break;
	default:
		check_mounted_device(device_name);
		break;
	}

	if ((device = alloc_device()) == NULL) {
		Err("structure allocation failed\n");
		exit(255);
	}

	if (open_device(device, device_name) == FAILED) {
		Err("unable to open %s\n", device->name);
		Info("%s\n", device->err_msg);
		exit(255);
	}

	if (!all_flg) {
		init_format_parameters(device);
	}

	if (chk_flg) {
		int status;

		status = check_device(device, ignore);
		show_device_info(device, status, verbose);
		exit(status);
	}

	if (!frc_flg && !device->is_dump_device) {
		Err("%s: initial format is required.\n", device->name);
		exit(NOT_A_DUMP_DEVICE);
	}

	if (!enough_block_size(device)) {
		Err("%s: %s %Lu\n", blksz_err, device->name,
					(u64)device->req->format_size);
	}

	show_format_info(device, verbose);

	device->do_progress = prg_flg;

	fmtd_blk = format_device(device);
	if (fmtd_blk == (off_t)-1) {
		Info("%s\n", device->reason);
		Err("formatting failed. (not suitable device)\n");
		exit(255);
	} else if (fmtd_blk != device->req->format_size) {
		Err("formatting failed. (block %Lu)\n", (u64)fmtd_blk);
		exit(255);
	} else {
		Info("formatting completed.\n");
		exit(FORMATTED);
	}
}
