/*
 *  diskdumpmsg.c
 *
 *  Copyright (C) 2007  FUJITSU LIMITED
 */
/* $Id: diskdumpmsg.c,v 1.3 2007/05/04 14:50:35 akira Exp $ */

#define _GNU_SOURCE

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <fcntl.h>
#include <time.h>
#include <zlib.h>
#include <errno.h>

#include "savecore.h"
#include "dumpfilter.h"
#include "diskdumpmsg.h"
#include "darray_string.h"
#include "systemmap.h"
#include "syslog.h"

/*
 * The following two functions are used for compressed vmcore
 */

static unsigned long pfn_to_block(DumpFile *dump, unsigned long pfn)
{
	unsigned long i, block = 0;

	for (i = 0; i < pfn; i++)
		if (page_is_dumpable(i, dump->device))
			block++;
	return block;
}

static int get_uncompressed_block(DumpFile *dump, void **buf)
{
	page_desc_t desc;
	int block_size = get_page_size();
	void *page_data = xmalloc(block_size);
	void *unc_page_data = xmalloc(block_size);

	if (read_page_desc(dump, &desc) != COMPLETE)
		return READ_ERROR_IN_PAGE_DESC;

	if (read_page_data(dump, page_data, &desc) != COMPLETE)
		return READ_ERROR_IN_PAGE_DATA;

	if (DumpPageCompressed(desc)) {
		unsigned long unc_len = block_size;
		if (uncompress(unc_page_data, &unc_len, page_data, desc.size)
				 != Z_OK || unc_len != block_size)
			return UNCOMPRESS_ERROR;

		free(page_data);
		*buf = unc_page_data;
	} else {
		free(unc_page_data);
		*buf = page_data;
	}

	return COMPLETE;
}

/*
 * The following two functions are used for ELF vmcore
 */

static off_t p_to_offset(DumpFile *dump, addr_t paddr)
{
	Elf_Phdr *phdr;

	darray_foreach(phdr, dump->loadsegments)
		if (phdr->p_paddr <= paddr &&
				paddr < phdr->p_paddr + phdr->p_memsz)
			return phdr->p_offset + (paddr - phdr->p_paddr);

	return (off_t)-1L;
}

static bool inline page_exists(DumpFile *dump, unsigned long pfn)
{
	addr_t paddr = pfn * get_page_size();

	return p_to_offset(dump, paddr) != (off_t)-1;
}

static int read_buffer(DumpFile *dump, addr_t vaddr, size_t len, void *buf)
{
	addr_t paddr;
	int block_size = get_page_size();
	unsigned long pfn;
	int ret;
	size_t copy_len, offs;
	void *page_data;

	paddr = vtop(dump, vaddr);
	pfn = paddr / block_size;
	offs = paddr % block_size;

	if (dump->type == DUMP_TYPE_COMPRESSED) {
		set_page_desc_offset(dump, pfn_to_block(dump, pfn));

		while (len > 0) {
			copy_len = min(len, block_size - offs);

			if (!page_is_dumpable(pfn, dump->device)) {
				memset(buf, 0, copy_len);
			} else {
				ret = get_uncompressed_block(dump, &page_data);
				if (ret != COMPLETE)
					return ret;

				memcpy(buf, page_data + offs, copy_len);

				free(page_data);
			}

			len -= copy_len;
			buf += copy_len;
			offs = 0;
			pfn++;
		}
	} else {
		while (len > 0) {
			copy_len = min(len, block_size - offs);

			if (!page_exists(dump, pfn)) {
				memset(buf, 0, copy_len);
			} else {
				ret = dump_pread(dump, buf, copy_len,
						p_to_offset(dump, paddr));
				if (ret != COMPLETE)
					return ret;
			}

			len -= copy_len;
			buf += copy_len;
			paddr += copy_len;
			offs = 0;
			pfn++;
		}
	}

	return COMPLETE;
}

static int read_buffer_sym(DumpFile *dump, const char *sym, size_t len, void *data)
{
	addr_t vaddr;

	vaddr = systemmap_vaddr(dump->map, sym);
	if (bad_addr(vaddr))
		return SYMBOL_ERROR;

	return read_buffer(dump, vaddr, len, data);
}

#define DEFINE_READ(TYPE)						\
									\
int read_##TYPE(DumpFile *dump, const char *sym, TYPE *val)		\
{									\
	return read_buffer_sym(dump, sym, sizeof(TYPE), val);		\
}

DEFINE_READ(ulong);
DEFINE_READ(int);
DEFINE_READ(long);

/* The maximum allowable size of log_buf */
static const int max_log_buf_len = 16777216;

static int get_logbuf(DumpFile *dump, char **_logbuf)
{
	char *logbuf;
	unsigned long log_start, log_end;
	addr_t log_buf;
	int log_buf_len;
	int len;
	int ret;

	ret = read_ulong(dump, "log_start", &log_start);
	if (ret != COMPLETE)
		return ret;
	ret = read_ulong(dump, "log_end", &log_end);
	if (ret != COMPLETE)
		return ret;
	ret = read_ulong(dump, "log_buf", &log_buf);
	if (ret != COMPLETE)
		return ret;
	ret = read_int(dump, "log_buf_len", &log_buf_len);
	if (ret != COMPLETE)
		return ret;

	if (log_buf_len > max_log_buf_len)
		return LOG_BUF_TOO_LARGE;

	logbuf = xmalloc(log_buf_len);

	if (log_end < log_buf_len) {
		ret = read_buffer(dump, log_buf, log_end, logbuf);
		if (ret != COMPLETE)
			goto err;

		len = log_end;
	} else {
		if (!log_buf_len) { 
			ret = READ_ERROR_IN_DUMP_FILE;
			goto err;
		}
		log_end %= log_buf_len;

		ret = read_buffer(dump, log_buf + log_end,
				log_buf_len - log_end, logbuf);
		if (ret != COMPLETE)
			goto err;
		ret = read_buffer(dump, log_buf, log_end,
				logbuf + log_buf_len - log_end);
		if (ret != COMPLETE)
			goto err;
		len = log_buf_len;
	}

err:
	if (ret != COMPLETE)
		free(logbuf);
	else
		*_logbuf = logbuf;

	return ret;
}

static bool
match_lines(darray_uint *da1, int idx1, darray_uint *da2, int idx2, int len)
{
	for (; len > 0; len--, idx1--, idx2--) {
		if (idx1 < 0 || idx2 < 0)
			return false;

		if (da1->ary[idx1] != da2->ary[idx2])
			return false;
	}

	return true;
}

static const int compare_lines = 3;

#define reverse_foreach(i, da, complines)	\
	for (i = (da)->length - 1; i >= (complines) - 1; i--)

/*
 * Return the number of the line from which the unrecorded messages
 * start in log_buf
 */
static int unrecorded_messages(darray_string *logbuf, darray_string *syslog)
{
	darray_uint *logbuf_hash, *syslog_hash;
	int lidx, sidx, ret = -1;

	logbuf_hash = darray_string_hash_array(logbuf);
	syslog_hash = darray_string_hash_array(syslog);

	reverse_foreach(lidx, logbuf_hash, compare_lines)
		reverse_foreach(sidx, syslog_hash, compare_lines)
			if (match_lines(logbuf_hash, lidx,
					syslog_hash, sidx, compare_lines)) {
				ret = lidx;
				goto out;
			}

out:
	darray_free(logbuf_hash);
	darray_free(syslog_hash);
	return ret;
}

static int setup_systemmap(DumpFile *dump)
{
	char *mapfile;
	darray_mapent *map = darray_mapent_alloc();

	darray_push(map, (mapent){ .name = "log_buf" });
	darray_push(map, (mapent){ .name = "log_buf" });
	darray_push(map, (mapent){ .name = "log_buf_len" });
	darray_push(map, (mapent){ .name = "log_start" });
	darray_push(map, (mapent){ .name = "log_end" });
	darray_push(map, (mapent){ .name = "xtime" });
	darray_push(map, (mapent){ .name = "system_utsname" });

	mapfile = xasprintf("/boot/System.map-%s", dump->utsname->release);

	if (systemmap_fill(map, mapfile) < 0)
		return FILE_OPEN_ERROR;

	dump->map = map;

	return COMPLETE;
}

int main(int argc, char *argv[])
{
	int opt;
	bool frc_flg = false, verbose = false, pipe_flg = false;
	const char *input = NULL;
	DumpFile *vmcore;
	char *logbuf;
	darray_string *logbuf_msgs, *syslog_msgs;
	int idx;
	int i;
	long timeofdeath;
	char timeofdeath_str[128];
	typeof(vmcore->utsname->nodename) hostname;
	FILE *fp;
	char *msg_file = VAR_LOG_MESSAGES;

	cmdname = "diskdumpmsg";

	while ((opt = getopt(argc, argv, "flpvV")) != -1) {
		switch (opt) {
		case 'f':
			frc_flg = true;
			break;
		case 'p':
			pipe_flg = true;
			break;
		case 'v':
			verbose = true;
			break;
		case 'V':
			version();
			break;
		default:
			usage();
			break;
		}
	}

	if (optind + 2 == argc)
		msg_file = argv[optind + 1];
	else if (optind + 1 != argc)
		usage();

	input = argv[optind];		/* get the source vmcore name */

	if (!does_exist(input)) {
		Err("%s does not exist\n", input);
		exit(1);
	}

	if (!does_exist(msg_file)) {
		Err("%s does not exists\n", msg_file);
		exit(1);
	}

	if ((vmcore = open_vmcore(input, EXISTING)) == NULL) {
		Err("unable to open %s\n", input);
		exit(1);
	}
	Info("input file\n");
	Info("  file name          : %s\n", vmcore->name);

	switch (read_vmcore_structs(vmcore)) {
	case COMPLETE:
		break;
	case NO_PANIC_DUMP_HEADER:
		Err("%s does not have panic dump header.\n", vmcore->name);
		exit(1);
	case READ_ERROR_IN_DUMP_HEADER:
		Err("unable to read the dump header in %s\n", vmcore->name);
		exit(1);
	case READ_ERROR_IN_DUMP_SUB_HEADER:
		Err("unable to read the dump sub header in %s\n",
								vmcore->name);
		exit(1);
	case BLOCK_SIZE_INVALID:
		Err("block size is invalid.\n");
		exit(1);
	case BLOCK_SIZE_NOT_MATCH:
		Err("block size of %s does not match with system page size.\n",
								vmcore->name);
		exit(1);
	case READ_ERROR_IN_MEMORY_BITMAP:
		Err("unable to read bitmap in %s\n", vmcore->name);
		exit(1);
	default:
		Err("unknown error (checking the dump header in %s)\n",
								vmcore->name);
		exit(1);
	}

	Info("  vmcore structs     : read\n");

	switch (setup_systemmap(vmcore)) {
	case COMPLETE:
		break;
	case FILE_OPEN_ERROR:
		Err("unable to open System.map-%s\n", vmcore->utsname->release);
		exit(1);
	default:
		Err("unknown error (reading System.map-%s)\n",
				vmcore->utsname->release);
		exit(1);
	}

	Info("  set up systemmap   : done\n");

	switch (get_logbuf(vmcore, &logbuf)) {
	case COMPLETE:
		break;
	case READ_ERROR_IN_PAGE_DESC:
		Err("unable to read descriptor (logbuf in %s)\n", vmcore->name);
		exit(1);
	case READ_ERROR_IN_PAGE_DATA:
		Err("unable to read data (logbuf in %s)\n", vmcore->name);
		exit(1);
	case SEEK_ERROR_IN_DUMP_FILE:
		Err("unable to seek (logbuf in %s)\n", vmcore->name);
		exit(1);
	case UNCOMPRESS_ERROR:
		Err("uncompress error (logbuf in %s)\n", vmcore->name);
		exit(1);
	case READ_ERROR_IN_DUMP_FILE:
		Err("unable to read (logbuf in %s)\n", vmcore->name);
		exit(1);
	case LOG_BUF_TOO_LARGE:
		Err("the length of log_buf is too large (logbuf in %s)\n",
				vmcore->name);
		exit(1);
	case SYMBOL_ERROR:
		Err("unknown symbol (logbuf in %s)\n", vmcore->name);
		exit(1);
	case INVALID_ADDRESS:
		Err("invalid address (logbuf in %s)\n", vmcore->name);
		exit(1);
	default:
		Err("unknown error (logbuf in %s)\n", vmcore->name);
		exit(1);
	}

	Info("  log_buf retrieving : done\n");

	switch (read_long(vmcore, "xtime", &timeofdeath)) {
	case COMPLETE:
		break;
	case READ_ERROR_IN_PAGE_DESC:
		Err("unable to read descriptor (xtime in %s)\n", vmcore->name);
		exit(1);
	case READ_ERROR_IN_PAGE_DATA:
		Err("unable to read data (xtime in %s)\n", vmcore->name);
		exit(1);
	case UNCOMPRESS_ERROR:
		Err("uncompress error (xtime in %s)\n", vmcore->name);
		exit(1);
	case READ_ERROR_IN_DUMP_FILE:
		Err("unable to read (xtime in %s)\n", vmcore->name);
		exit(1);
	default:
		Err("unknown error (xtime in %s)\n", vmcore->name);
		exit(1);
	}

	strftime(timeofdeath_str, sizeof(timeofdeath_str),
			"%b %d %H:%M:%S", localtime(&timeofdeath));

	Info("  time of death      : %s\n", timeofdeath_str);

	if (!frc_flg && !pipe_flg
			&& is_later(msg_file, timeofdeath)) {
		Err("messages are later than vmcore\n");
		exit(1);
	}

	strncpy(hostname, vmcore->utsname->nodename, sizeof(hostname) - 1);
	chop(hostname, ".");

	Info("  hostname           : %s\n", hostname);

	logbuf_msgs = split_regex(logbuf, "\n(<[1234567890]+>)?", 0);
	syslog_msgs = last_kernel_messages(msg_file);

	idx = unrecorded_messages(logbuf_msgs, syslog_msgs);

	Info("output\n");

	if (pipe_flg) {
		if ((fp = fdopen(1, "w")) == NULL) {
			Err("unable to open stdout\n");
			exit(1);
		}
		Info("  send messages to   : stdout\n");
	} else {
		if ((fp = fopen(msg_file, "a")) == NULL) {
			Err("unable to open %s\n", msg_file);
			exit(1);
		}
		Info("  send messages to   : %s\n", msg_file);
	}

	fprintf(fp, "%s %s " SALV_HEADER "\n", timeofdeath_str, hostname);

	for (i = idx + 1; i < logbuf_msgs->length; i++)
		fprintf(fp, "%s %s kernel: %s\n", timeofdeath_str,
				hostname, logbuf_msgs->ary[i]);

	fprintf(fp, "%s %s " SALV_FOOTER "\n", timeofdeath_str, hostname);

	fclose(fp);

	darray_clear(logbuf_msgs);
	darray_clear(syslog_msgs);

	Info("message complement completed.\n");
	exit(0);
}
