/*
 *  dumpfile.c
 *
 *  Copyright (C) 2005-2007  FUJITSU LIMITED
 *
 */
/* $Id: dumpfile.c,v 1.10 2007/05/04 14:50:35 akira Exp $ */

#define _GNU_SOURCE

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <time.h>
#include "savecore.h"

DumpFile *open_dump_file(char *dir)
{
	DumpFile *dump_file = xcalloc(1, sizeof(DumpFile));
	char *filename = xasprintf("%s%s", dir, INCOMPLETE_DUMP_FILE);
	int fd;

	dump_file->dir = dir;

	fd = open(filename, O_WRONLY|O_TRUNC|O_CREAT, S_IRUSR|S_IWUSR);
	if (fd < 0) {
		free(dump_file);
		free(filename);
		return NULL;
	}

	dump_file->fd	= fd;
	dump_file->name	= filename;

	return dump_file;
}

DumpFile *open_vmcore(const char *name, int flag)
{
	DumpFile *vmcore = xcalloc(1, sizeof(*vmcore));

	vmcore->device = xcalloc(1, sizeof(DumpDevice));

	switch (flag) {
	case NEW:
		if ((vmcore->fd = open(name,
			O_WRONLY|O_TRUNC|O_CREAT, S_IRUSR|S_IWUSR)) < 0) {
			goto err;
		}
		break;

	case EXISTING:
	default:
		if ((vmcore->fd = open(name, O_RDONLY)) < 0) {
			goto err;
		}
		break;
	}

	vmcore->dir = "";
	vmcore->name = name;

	return vmcore;
err:
	free(vmcore->device);
	free(vmcore);

	return NULL;
}

int rename_dump_file(DumpFile *dump_file)
{
	char *old_file = xasprintf("%s%s",
					dump_file->dir, INCOMPLETE_DUMP_FILE);
	char *new_file = xasprintf("%s%s", dump_file->dir, COMPLETE_DUMP_FILE);
	int ret;

	ret = rename(old_file, new_file);
	dump_file->name = xstrdup(new_file);

	free(old_file);
	free(new_file);

	return ret;
}

int read_page_desc(DumpFile *dump_file, page_desc_t *desc)
{
	if (dump_seek(dump_file, dump_file->page_desc_offset) < 0)
		return SEEK_ERROR_IN_DUMP_DEVICE;

	if (read(dump_file->fd, desc, sizeof(*desc)) != sizeof(*desc))
		return READ_ERROR_IN_DUMP_FILE;

	dump_file->page_desc_offset += sizeof(*desc);

	return COMPLETE;
}

int write_page_desc(DumpFile *dump_file, page_desc_t *desc)
{
	if (dump_seek(dump_file, dump_file->page_desc_offset) < 0)
		return SEEK_ERROR_IN_DUMP_DEVICE;

	if (write(dump_file->fd, desc, sizeof(*desc)) != sizeof(*desc))
		return WRITE_ERROR_IN_DUMP_FILE;

	dump_file->page_desc_offset += sizeof(*desc);

	return COMPLETE;
}

int write_page_data(DumpFile *dump_file, const void *data, page_desc_t *desc)
{
	if (dump_seek(dump_file, desc->offset) < 0)
		return SEEK_ERROR_IN_DUMP_DEVICE;

	if (write(dump_file->fd, data, desc->size) != desc->size)
		return WRITE_ERROR_IN_DUMP_FILE;

	dump_file->data_offset = desc->offset + desc->size;

	return COMPLETE;
}

int read_page_data(DumpFile *dump_file, void *data, page_desc_t *desc)
{
	if (dump_seek(dump_file, desc->offset) < 0)
		return SEEK_ERROR_IN_DUMP_DEVICE;

	if (read(dump_file->fd, data, desc->size) != desc->size)
		return READ_ERROR_IN_DUMP_FILE;

	dump_file->data_offset += desc->size;

	return COMPLETE;
}

int dump_read(DumpFile *dump_file, void *data, uint32_t size)
{
	if (dump_seek(dump_file, dump_file->data_offset) < 0)
		return SEEK_ERROR_IN_DUMP_DEVICE;

	if (read(dump_file->fd, data, size) != size)
		return READ_ERROR_IN_DUMP_FILE;

	dump_file->data_offset
		+= roundup(size, dump_file->device->header->block_size);

	return COMPLETE;
}

int dump_write(DumpFile *dump_file, const void *data, uint32_t size)
{
	if (dump_seek(dump_file, dump_file->data_offset) < 0)
		return SEEK_ERROR_IN_DUMP_DEVICE;

	if (write(dump_file->fd, data, size) != size)
		return WRITE_ERROR_IN_DUMP_FILE;

	dump_file->data_offset
		+= roundup(size, dump_file->device->header->block_size);

	return COMPLETE;
}

int dump_pread(DumpFile *dump_file, void *data, uint32_t size, off_t off)
{
	if (dump_seek(dump_file, off) < 0)
		return SEEK_ERROR_IN_DUMP_DEVICE;

	if (read(dump_file->fd, data, size) != size)
		return READ_ERROR_IN_DUMP_FILE;

	return COMPLETE;
}

void *alloc_blocks(DumpFile *dump_file, int block_nr)
{
	return xcalloc(1, dump_file->device->header->block_size * block_nr);
}

int read_blocks(DumpFile *dump_file, void *data, int block_nr)
{
	return dump_read(dump_file, data,
			 block_nr * dump_file->device->header->block_size);
}

int write_blocks(DumpFile *dump_file, const void *data, int block_nr)
{
	return dump_write(dump_file, data,
			  block_nr * dump_file->device->header->block_size);
}

void set_data_offset(DumpFile *dump_file)
{
	dump_file->data_offset = dump_file->device->header->block_size
				* (1 + dump_file->device->header->sub_hdr_size);
}

void set_page_desc_offset(DumpFile *dump_file, unsigned long idx)
{
	dump_file->page_desc_offset = dump_file->page_desc_offset_base +
					idx * sizeof(page_desc_t);
}

void calc_data_offset(DumpFile *dump_file)
{
	int i;
	unsigned long num_pages = 0;

	for (i = 0; i <= dump_file->device->header->max_mapnr; i++)
		if (page_is_dumpable(i, dump_file->device))
			num_pages++;
	dump_file->data_offset += num_pages * sizeof(page_desc_t);
}

DEFINE_DARRAY(elfphdr);

static void read_load_segments(DumpFile *src, char *header)
{
	Elf_Ehdr *elf;
	Elf_Phdr *phdr;
	darray_elfphdr *phdrs = darray_elfphdr_alloc();
	int i;

	elf = (typeof(elf))header;

	phdr = (typeof(phdr))(header + elf->e_phoff);
	for (i = 0; i < elf->e_phnum; i++) {
		phdr = (typeof(phdr))((char *)phdr + elf->e_phentsize);

		if (phdr->p_type == PT_LOAD)
			darray_push(phdrs, *phdr);
	}

	src->loadsegments = phdrs;
}

static bool inline is_elf_vmcore(const char *header)
{
	return memcmp(header, ELFMAG, SELFMAG) == 0;
}

static int read_utsname(int fd, struct new_utsname **un)
{
	int ret;
	unsigned int page_size = get_page_size();
	void *page_buf_base = xcalloc(2, page_size);
	void *page_buf = page_buf_base + page_size;

	lseek(fd, 0, SEEK_SET);
	for (;;) {
		typeof((*un)->sysname) name = "Linux";
		int namesz = sizeof(name);
		void *p;

		ret = read(fd, page_buf, page_size);
		if (ret < 0) {
			ret = READ_ERROR_IN_DUMP_FILE;
			break;
		} else if (ret == 0) {
			ret = UTSNAME_NOT_FOUND;
			break;
		}

		if ((p = memmem(page_buf - namesz, ret + namesz, name,
				namesz)) != NULL) {
			*un = memdup((typeof(*un))p);
			ret = COMPLETE;
			break;
		}

		memcpy(page_buf - namesz, page_buf + page_size - namesz,
			namesz);
	}

	free(page_buf_base);
	return ret;
}

static int read_vmcore_bitmap(DumpFile *src)
{
	src->device->bitmap = alloc_blocks(src,
				src->device->header->bitmap_blocks);

	set_data_offset(src);

	if (read_blocks(src, src->device->bitmap,
			src->device->header->bitmap_blocks) != COMPLETE) {
		return READ_ERROR_IN_MEMORY_BITMAP;
	}

	src->device->dumpable_bitmap = get_dumpable_bitmap(src->device);

	src->page_desc_offset_base = src->data_offset;
	src->page_desc_offset = src->data_offset;

	calc_data_offset(src);

	return COMPLETE;
}

static int read_elf_vmcore(DumpFile *src, void *header)
{
	src->type = DUMP_TYPE_ELF;

	read_load_segments(src, header);
	free(header);

	if (src->loadsegments->length < 1) {
		return INVALID_ELF_SEGMENTS;
	}

	return read_utsname(src->fd, &src->utsname);
}

static int read_compressed_vmcore(DumpFile *src,
					struct disk_dump_header *header)
{
	struct disk_dump_sub_header *sub_header;
	const int block_size = get_page_size();

	src->type = DUMP_TYPE_COMPRESSED;

	if (!dump_signature_is_good(header->signature)) {
		free(header);
		return NO_PANIC_DUMP_HEADER;
	}
	if (header->block_size != block_size) {
		free(header);
		return BLOCK_SIZE_NOT_MATCH;
	}

	src->device->header = header;

	/* read vmcore sub header */
	if ((sub_header = read_alloc(src->fd, block_size)) == NULL) {
		return READ_ERROR_IN_DUMP_SUB_HEADER;
	}

	src->device->sub_header = sub_header;

	src->device->bitmap_len = block_size * header->bitmap_blocks;

	src->utsname = &header->utsname;

	return read_vmcore_bitmap(src);
}

int read_vmcore_structs(DumpFile *src)
{
	void *header;

	/* read the first block */
	if ((header = read_alloc(src->fd, get_page_size())) == NULL) {
		return READ_ERROR_IN_DUMP_HEADER;
	}

	if (is_elf_vmcore(header))
		return read_elf_vmcore(src, header);
	else
		return read_compressed_vmcore(src, header);
}

void show_header_info(struct disk_dump_header *hdr, bool verbose)
{
	Info("header info\n");
	Info("  total device blocks: %d\n", hdr->device_blocks);
	Info("  bitmap length      : %d\n", hdr->bitmap_blocks);
	Info("  max map number     : %d\n", hdr->max_mapnr);
	Info("  total RAM blocks   : %d\n", hdr->total_ram_blocks);
	Info("  written blocks     : %d\n", hdr->written_blocks);
}

void show_dump_info(struct disk_dump_header *hdr, bool verbose)
{
	const char *timeofdeath	= get_date_time_str(hdr->timestamp.tv_sec);
	const char *current	= get_date_time_str(CURRENT_DATE_TIME);

	Info("dump info\n");
	Info("  kernel             : %s\n", kernel_path(hdr->utsname.release));
	Info("  node name          : %s\n", hdr->utsname.nodename);
	Info("  release            : %s\n", hdr->utsname.release);
	Info("  machine            : %s\n", hdr->utsname.machine);
	Info("  dumped date        : %s\n", timeofdeath);
	Info("  current date       : %s\n", current);
	Info("  passed days        : %d\n",
			diff_days(CURRENT_DATE_TIME, hdr->timestamp.tv_sec));
}
