/*
 *  dumpfilter.c
 *
 *  Copyright (C) 2006, 2007  FUJITSU LIMITED
 *
 */
/* $Id: dumpfilter.c,v 1.12 2007/05/18 15:35:44 akira Exp $ */

#define _GNU_SOURCE

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <fcntl.h>
#include <time.h>

#include "savecore.h"
#include "dumpfilter.h"

#define CMD	"dumpfilter"

static int dump_level = DEFAULT_DUMP_LEVEL;
static int skipped_page_cnt	= 0;
static int saved_page_cnt	= 0;
static bool prog_flg = false;	/* becomes true if option '-p' specified. */
static const int symbol = '#';	/* for progress bar */

static int write_vmcore_header(DumpFile *dst, DumpFile *src)
{
	/* save original data_offset */
	const off_t data_offset = dst->data_offset;

	/* write vmcore header */
	dst->data_offset = 0;
	if (write_blocks(dst, src->device->header, 1) != COMPLETE) {
		return WRITE_ERROR_IN_DUMP_HEADER;
	}

	/* write vmcore sub header */
	if (write_blocks(dst, src->device->sub_header, 1) != COMPLETE) {
		return WRITE_ERROR_IN_DUMP_SUB_HEADER;
	}

	dst->data_offset = data_offset;

	return COMPLETE;
}

static void dup_vmcore(DumpFile *dst, DumpFile *src)
{
	dst->device->header		= src->device->header;
	dst->device->sub_header		= src->device->sub_header;
	dst->device->sub_header_len	= src->device->sub_header_len;
	dst->device->bitmap_len		= src->device->bitmap_len;

	dst->data_offset		= src->data_offset;
	dst->page_desc_offset		= src->page_desc_offset;

	dst->device->bitmap = alloc_blocks(dst,
				dst->device->header->bitmap_blocks);

	memcpy(dst->device->bitmap, src->device->bitmap,
			dst->device->bitmap_len);

	dst->device->dumpable_bitmap = get_dumpable_bitmap(dst->device);
}

static int write_vmcore_bitmap(DumpFile *dst)
{
	set_data_offset(dst);
	if (write_blocks(dst, dst->device->bitmap,
			dst->device->header->bitmap_blocks) != COMPLETE) {
		return FAILED;
	}

	return SUCCEEDED;
}

static inline int PageAnon(page_desc_t *pd)
{
	return pd->flags & DUMP_DH_MAPPING_ANON;
}

/*
 * determine whether to save the page or not.
 *	true : page is saved
 *	false: page is NOT saved
 */
static bool page_saved(page_desc_t *pd)
{
	if (!dump_level) {
		return true;
	}
	if ((dump_level & DUMP_EXCLUDE_CACHE) &&
	    (dump_level & DUMP_SAVE_PRIVATE) && PagePrivate(pd)) {
		return true;
	}
	if ((dump_level & DUMP_EXCLUDE_CACHE) && !PageAnon(pd) &&
	    (PageLRU(pd) || PageSwapCache(pd))) {
		return false;
	}
	if ((dump_level & DUMP_EXCLUDE_ANON) && PageAnon(pd)) {
		return false;
	}

	return true;
}

static int write_vmcore_pages(DumpFile *dst, DumpFile *src)
{
	int i;
	char *data;
	page_desc_t pd;

	data = alloc_blocks(dst, 1);

	for (i = 0; i <= src->device->header->max_mapnr; i++) {
		if (prog_flg) {
			progress_bar(dst->name, symbol, i,
					src->device->header->max_mapnr);
		}

		if (!page_is_dumpable(i, src->device)) {
			continue;
		}

		if (read_page_desc(src, &pd) != COMPLETE) {
			free(data);
			return READ_ERROR_IN_PAGE_DESC;
		}

		if (!page_saved(&pd)) {
			page_dumpable_clear(i, dst->device);
			skipped_page_cnt++;
			continue;
		}

		saved_page_cnt++;

		if (read_page_data(src, data, &pd) != COMPLETE) {
			free(data);
			return READ_ERROR_IN_PAGE_DATA;
		}

		pd.offset = dst->data_offset;

		if (write_page_desc(dst, &pd) != COMPLETE) {
			free(data);
			return WRITE_ERROR_IN_PAGE_DESC;
		}

		if (write_page_data(dst, data, &pd) != COMPLETE) {
			free(data);
			return WRITE_ERROR_IN_PAGE_DATA;
		}
	}
	if (prog_flg) {
		progress_bar_post(dst->name, symbol, i,
				src->device->header->max_mapnr);
	}

	return COMPLETE;
}

static inline void get_dump_level(char *val)
{
	dump_level = strtol(val, NULL, 0);
}

int main(int argc, char *argv[])
{
	int opt;
	bool frc_flg = false, verbose = false;
	const char *input = NULL, *output = NULL;
	DumpFile *vmcore_i, *vmcore_o;

	cmdname = CMD;

	while ((opt = getopt(argc, argv, "fl:pvV")) != -1) {
		switch (opt) {
		case 'f':
			frc_flg = true;
			break;
		case 'l':
			get_dump_level(optarg);
			break;
		case 'p':
			prog_flg = true;
			break;
		case 'v':
			verbose = true;
			break;
		case 'V':
			version();
			break;
		default:
			usage();
			break;
		}
	}

	if (optind + 2 != argc) {
		usage();
	}

	input = argv[optind];		/* get the source vmcore name */
	output = argv[optind + 1];	/* get the target vmcore name */

	if (!does_exist(input)) {
		Err("%s does not exist\n", input);
		exit(1);
	}

	if (does_exist(output)) {
		if (!frc_flg) {
			Err("%s already exists\n", output);
			exit(1);
		}
		if (same_file(input, output)) {
			Err("source and target are identical\n");
			exit(1);
		}
	}

	Info("filtering\n");
	Info("  dump level         : %d\n", dump_level);
	if (frc_flg) {
		Info("  note               : %s will be overwritten\n", output);
	}

	if ((vmcore_i = open_vmcore(input, EXISTING)) == NULL) {
		Err("unable to open %s\n", input);
		exit(1);
	}
	Info("input file\n");
	Info("  file name          : %s\n", vmcore_i->name);

	switch (read_vmcore_structs(vmcore_i)) {
	case COMPLETE:
		break;
	case NO_PANIC_DUMP_HEADER:
		Err("%s does not have panic dump header.\n", vmcore_i->name);
		exit(1);
	case READ_ERROR_IN_DUMP_HEADER:
		Err("unable to read the dump header in %s\n", vmcore_i->name);
		exit(1);
	case READ_ERROR_IN_DUMP_SUB_HEADER:
		Err("unable to read the dump sub header in %s\n",
								vmcore_i->name);
		exit(1);
	case BLOCK_SIZE_INVALID:
		Err("block size is invalid.\n");
		exit(1);
	case BLOCK_SIZE_NOT_MATCH:
		Err("block size of %s does not match with system page size.\n",
								vmcore_i->name);
		exit(1);
	case READ_ERROR_IN_MEMORY_BITMAP:
		Err("unable to read bitmap in %s\n", vmcore_i->name);
		exit(1);
	default:
		Err("unknown error (checking the dump header in %s)\n",
								vmcore_i->name);
		exit(1);
	}
	Info("  vmcore structure      : checked\n");

	if (vmcore_i->type != DUMP_TYPE_COMPRESSED) {
		Err("%s is not a compressed vmcore\n", vmcore_i->name);
		exit(1);
	}

	show_header_info(vmcore_i->device->header, verbose);
	show_dump_info(vmcore_i->device->header, verbose);

	if ((vmcore_o = open_vmcore(output, NEW)) == NULL) {
		Err("unable to open %s\n", output);
		exit(1);
	}
	Info("output file\n");
	Info("  file name          : %s\n", vmcore_o->name);

	dup_vmcore(vmcore_o, vmcore_i);

	switch (write_vmcore_header(vmcore_o, vmcore_i)) {
	case COMPLETE:
		break;
	case WRITE_ERROR_IN_DUMP_HEADER:
		Err("unable to write data to header.\n");
		goto err;
	case WRITE_ERROR_IN_DUMP_SUB_HEADER:
		Err("unable to write data to sub header.\n");
		goto err;
	default:
		Err("unknown error (writing to the dump header)\n");
		goto err;
	}
	Info("  vmcore header      : written\n");

	if (!prog_flg) {
		Info("  vmcore pages       : ");
	}
	switch (write_vmcore_pages(vmcore_o, vmcore_i)) {
	case COMPLETE:
		break;
	case READ_ERROR_IN_PAGE_DESC:
		Err("unable to read page descriptor\n");
		goto err;
	case READ_ERROR_IN_PAGE_DATA:
		Err("unable to read page data\n");
		goto err;
	case WRITE_ERROR_IN_PAGE_DESC:
		Err("unable to write page descriptor\n");
		goto err;
	case WRITE_ERROR_IN_PAGE_DATA:
		Err("unable to write page data\n");
		goto err;
	default:
		Err("unknown error (writing to the vmcore pages)\n");
		goto err;
	}
	if (!prog_flg) {
		Info("written\n");
	} else {
		Info("  vmcore pages       : written\n");
	}

	if (write_vmcore_bitmap(vmcore_o) == FAILED) {
		Err("unable to write vmcore bitmap to %s\n", vmcore_o->name);
		goto err;
	}
	Info("  vmcore bitmap      : written\n");

	Info("  skipped pages      : %d\n", skipped_page_cnt);
	Info("  saved pages        : %d\n", saved_page_cnt);

	Info("dump filtering completed.\n");

	exit(0);
err:

	if (unlink(vmcore_o->name) < 0) {
		Err("unable to delete %s\n", vmcore_o->name);
	}
	exit(1);
}
