/*
 *  savecore.c
 *
 *  Copyright (C) 2004-2007  FUJITSU LIMITED
 *  Copyright (C) 2001-2005  Red Hat, Inc
 *
 *  Written by Nobuhiro Tachino (tachino@jp.fujitsu.com) based on
 *  netdump-server code.
 */
/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */
/* $Id: savecore.c,v 1.82 2007/05/31 15:33:47 akira Exp $ */

#define _GNU_SOURCE
#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <fcntl.h>
#include <sys/statfs.h>
#include <errno.h>
#include <string.h>
#include <zlib.h>

#include "savecore.h"

#define CMD	"savecore"

#ifdef FORCE_SINGLE_PTLOAD
#define _FORCE_SINGLE_PTLOAD	true
#else
#define _FORCE_SINGLE_PTLOAD	false
#endif

static bool prog_flg = false;	/* becomes true if option '-p' specified. */
static bool frc_flg = false;	/* becomes true if option '-f' specified. */
static unsigned long long pagetype_flg; /* if matches, page is uncompressed */

static const int symbol = '#';	/* for progress bar */

bool generate_elf_header(DumpDevice *);


static DumpDevice *create_new_device(const char *device_name)
{
	DumpDevice *device = xcalloc(1, sizeof(DumpDevice));

	device->name = xstrdup(device_name);

	device->fd = -1;

	/* Lame default. should be read from kernel */
	device->page_offset = PAGE_OFFSET;

	return device;
}

static int open_device_file(DumpDevice *device)
{
	int fd;

	if ((fd = open(device->name, O_RDONLY)) < 0)
		return FAILED;

	device->fd = fd;

	return SUCCEEDED;
}

static int reopen_device_file(DumpDevice *device)
{
	int fd;

	if (close(device->fd) < 0)
		return FAILED;
	if ((fd = open(device->name, O_RDWR)) < 0)
		return FAILED;

	device->fd = fd;

	return SUCCEEDED;
}

static int check_fs(DumpDevice *device)
{
	struct statfs buf;
	off_t memdump_size;
	bool tried_once = false;
	enum {
		available,
		no_space
	};
	off_t fs_size;

redo:
	if (fstatfs(device->dump_file->fd, &buf) < 0)
		return FAILED;

	/* elf header */
	memdump_size = (off_t)device->header->written_blocks + 3;
	memdump_size *= device->header->block_size;
	memdump_size = (memdump_size + (buf.f_bsize - 1)) / buf.f_bsize;

	if (is_super_user() && !same_fs(device->dump_file->dir, "/"))
		fs_size = buf.f_bfree;
	else
		fs_size = buf.f_bavail;

	if (memdump_size <= fs_size)
		return available;

	if (!tried_once++)
		if (execute_script(DISKDUMP_NOSPACE, device->dump_file->dir)
				> 0)
			goto redo;

	return no_space;
}

static char *make_dump_dir(DumpDevice *device)
{
	char *path;
	const char *ip_addr = convert_ip_addr(DISKDUMP_IP);
	char *date = get_date_time_str(device->header->timestamp.tv_sec);

	if (ip_addr == NULL || date == NULL) {
		return NULL;
	}
	path = xasprintf("%s%s-%s/", DISKDUMP_DIR_PREFIX, ip_addr, date);

	if (frc_flg && rm_dir(path) < 0) {
		free(path);
		return NULL;
	}

	if (mkdir(path, S_IRWXU) < 0) {
		free(path);
		return NULL;
	}

	return path;
}

static inline int get_bit(char *map, int byte, int bit)
{
	return map[byte] & (1<<bit);
}

static inline void set_bit(char *map, int byte, int bit)
{
	map[byte] |= 1<<bit;
}

static int next_bit(DumpDevice *device)
{
	char *src = get_dumpable_bitmap(device);
	int val;

	val = get_bit(src, device->byte, device->bit);
	if (++device->bit == 8) {
		device->byte++;
		device->bit = 0;
	}

	return val;
}

static void expand_bitmap(DumpDevice *device)
{
	int byte, bit;

	for (byte = 0; byte < device->bitmap_len / 2; byte++)
		for (bit = 0; bit < 8; bit++)
			if (get_bit(device->bitmap, byte, bit)
			 		&& next_bit(device))
				set_bit(device->dumpable_bitmap, byte, bit);
}

static int read_device_header(DumpDevice *device)
{
	struct disk_dump_header *header;
	struct disk_dump_sub_header *sub_header;
	int ret;
	int bitmap_len;
	const int block_size = get_page_size();
	int sub_header_len;
	off_t offset;

	header = xmalloc(block_size);

	offset = (off_t)block_size;
	if (device_seek(device, offset) < 0) {
		ret = SEEK_ERROR_IN_DUMP_HEADER;
		goto err;
	}

	if (read(device->fd, header, block_size) < block_size) {
		ret = READ_ERROR_IN_DUMP_HEADER;
		goto err;
	}

	/* validate dump header */
	if (!dump_signature_is_good(header->signature)) {
		ret = NO_PANIC_DUMP_HEADER;
		goto err;
	}

	if (header->block_size != block_size) {
		ret = BLOCK_SIZE_NOT_MATCH;
		goto err;
	}

	if (sizeof(*header) + sizeof(void *) * header->nr_cpus > block_size
			|| header->nr_cpus <= 0) {
		ret = NUMBER_OF_CPUS_INVALID;
		goto err;
	}

	/* read sub header */
	offset = (off_t)block_size * 2;
	if (device_seek(device, offset) < 0) {
		ret = SEEK_ERROR_IN_DUMP_SUB_HEADER;
		goto err;
	}

	sub_header_len	= sub_header_length(header);
	sub_header	= xmalloc(sub_header_len);

	if (read(device->fd, sub_header, sub_header_len) < sub_header_len) {
		ret = READ_ERROR_IN_DUMP_SUB_HEADER;
		goto err;
	}

	device->sub_header	= sub_header;
	device->sub_header_len	= sub_header_len;

	/* read memory bitmap */
	bitmap_len		= block_size * header->bitmap_blocks;
	device->bitmap_len	= bitmap_len;

	offset = (off_t)block_size * (2 + header->sub_hdr_size);
	if (device_seek(device, offset) < 0) {
		ret = SEEK_ERROR_IN_MEMORY_BITMAP;
		goto err;
	}

	device->bitmap		= xmalloc(bitmap_len);
	device->dumpable_bitmap	= xcalloc(1, bitmap_len);

	if (read(device->fd, device->bitmap, bitmap_len) < bitmap_len) {
		ret = READ_ERROR_IN_MEMORY_BITMAP;
		goto err;
	}

	if (dump_is_partial(header))
		expand_bitmap(device);
	else
		memcpy(device->dumpable_bitmap, device->bitmap, bitmap_len);

	/*
	 * page 0: device header
	 * page 1: dump header
	 * page 2..: dump sub header
	 * page 2 + size of subheader: memory page 0
	 */
	device->device_page_start
		= (2 + header->sub_hdr_size + header->bitmap_blocks)
		* header->block_size;

	device->header = header;

	return COMPLETE;

err:
	free(header);
	return ret;
}

static bool pages_are_zero(int nr, int len, DumpDevice *device)
{
	int i;
	const unsigned int maxnr = nr + min(len, device->header->max_mapnr);

	for (i = nr >> 3; i < (maxnr + 7) >> 3; i++)
		if (device->bitmap[i])
			return false;

	return true;
}

static void build_segments_info(DumpDevice *device)
{
	struct mem_segment *seg, *seg_base;
	unsigned int chunk;	/* = HOLE_SIZE blocks */
	const unsigned int max_segments
				= device->header->max_mapnr / HOLE_SIZE / 2;
	bool in_segment;

	seg = seg_base = xcalloc(max_segments, sizeof(struct mem_segment));

	if (_FORCE_SINGLE_PTLOAD || device->header->max_mapnr <= HOLE_SIZE) {
		seg->blk = 0;
		seg->len = device->header->max_mapnr;

		device->num_segments = 1;
		device->segments = seg_base;

		return;
	}

	/* The first segment must have the first block(blk#0). */
	seg = seg_base;
	seg->blk = 0;
	in_segment = true;
	for (chunk = 1; chunk < divideup(device->header->max_mapnr, HOLE_SIZE);
			chunk++) {
		if (pages_are_zero(chunk * HOLE_SIZE, HOLE_SIZE, device)) {
			if (in_segment) {
				seg->len = chunk * HOLE_SIZE - seg->blk;
				in_segment = false;
			}
		} else {
			if (!in_segment) {
				seg++;
				seg->blk = chunk * HOLE_SIZE;
				in_segment = true;
			}
		}
	}
	/*
	 * The last segment must contain the last blocks even if
	 * they are zero-filled pages.
	 */
	seg->len = device->header->max_mapnr - seg->blk;

	device->num_segments = seg + 1 - seg_base;
	device->segments = seg_base;
}

static bool page_is_in_segs(unsigned int nr, DumpDevice *device)
{
	struct mem_segment *seg;
	int i;

	for (i = 0; i < device->num_segments; i++) {
		seg = &device->segments[i];

		if (seg->blk <= nr && nr < seg->blk + seg->len)
			return true;
	}

	return false;
}

/******************************************************************************
 *                       compressed dump                                      *
 ******************************************************************************/
static inline bool dump_is_compressed(DumpDevice *device)
{
	return device->header->status & DUMP_HEADER_COMPRESSED;
}

static inline bool write_compress_header(DumpDevice *device)
{
	/* write dump header */
	device->dump_file->data_offset = 0;
	if (dump_write(device->dump_file, device->header,
			device->header->block_size) != COMPLETE)
		return FAILED;

	/* write dump sub header */
	if (dump_write(device->dump_file, device->sub_header,
			device->sub_header_len) != COMPLETE)
		return FAILED;

	/* write bitmap */
	if (dump_is_partial(device->header)) {
		int size = device->bitmap_len / 2;
		/* write 1st-bitmap */
		if (dump_write(device->dump_file, device->bitmap, size)
				!= COMPLETE)
			return FAILED;
		if (dump_write(device->dump_file, device->dumpable_bitmap, size)
				!= COMPLETE)
			return FAILED;
	} else {
		if (dump_write(device->dump_file, device->bitmap,
				device->bitmap_len) != COMPLETE)
			return FAILED;
	}

	device->dump_file->page_desc_offset = device->dump_file->data_offset;

	calc_data_offset(device->dump_file);

	return SUCCEEDED;
}

static int write_compress_pages(DumpDevice *device, unsigned char *scratch[])
{
	struct dump_page_header ph;
	int ret;
	unsigned long len;
	unsigned char *rawdata = scratch[0], *data = scratch[1];
	page_desc_t desc;

	/* read page header */
	if (read(device->fd, &ph, sizeof(ph)) != sizeof(ph))
		return READ_ERROR_IN_DUMP_DEVICE;

	/* sanity check */
	if (ph.size > device->header->block_size)
		return INVALID_SIZE_IN_PAGE_HEADER;

	/* try to read in the dump page itself */
	if (read(device->fd, rawdata, ph.size) < ph.size)
		/* if this fails, we may have run out of space */
		return READ_ERROR_IN_DUMP_DEVICE;

	/* write out the dump page header */
	if (device->dump_file->page_desc_offset
			> device->dump_file->data_offset)
		return SEEK_ERROR_IN_DUMP_DEVICE;

	desc.offset	= device->dump_file->data_offset;
	desc.page_flags	= ph.page_flags;
	if (DumpPageCompressed(ph) && ph.page_flags & pagetype_flg) {
		desc.size	= device->header->block_size;
		desc.flags	= ph.flags & ~DUMP_DH_COMPRESSED;

		ret = write_page_desc(device->dump_file, &desc);
		if (ret != COMPLETE)
			return ret;

		len = device->header->block_size;
		ret = uncompress(data, &len, rawdata, ph.size);
		if (ret != Z_OK || len != device->header->block_size)
			return UNCOMPRESS_ERROR;

		ret = write_page_data(device->dump_file, data, &desc);
	} else {
		desc.size	= ph.size;
		desc.flags	= ph.flags;

		ret = write_page_desc(device->dump_file, &desc);
		if (ret != COMPLETE)
			return ret;

		/* try to write out the dump page itself */
		ret = write_page_data(device->dump_file, rawdata, &desc);
	}
	if (ret != COMPLETE)
		return ret;

	return COMPLETE;
}

/******************************************************************************
 *                       Elf core dumping                                     *
 ******************************************************************************/

static inline int write_dump_header(DumpDevice *device)
{
	if (dump_is_compressed(device))
		return write_compress_header(device);

	return generate_elf_header(device) ? FAILED : SUCCEEDED;
}

static inline bool check_zero_page(DumpDevice *device, unsigned char *data)
{
	int i;

	for (i = 0; i < device->header->block_size; i++)
		if (data[i])
			return false;

	return true;
}

static inline bool page_is_ram(unsigned int nr, DumpDevice *device)
{
	return get_bit(device->bitmap, nr >> 3, nr & 7);
}

static inline bool dump_is_short_area(DumpDevice *device)
{
	return (device->header->status == DUMP_HEADER_SHORT_AREA);
}

static int write_pages(DumpDevice *device)
{
	int i;
	unsigned char *data[2] = {NULL, NULL};
	int ret = COMPLETE;
	const int block_size = device->header->block_size;
	off_t offset;
	const off_t failed = (off_t)-1;

	data[0] = xmalloc(block_size);
	data[1] = xmalloc(block_size);

	offset = (off_t)device->device_page_start;
	if (device_seek(device, offset) < 0) {
		ret = SEEK_ERROR_IN_DUMP_DEVICE;
		goto out;
	}

	offset = device->dump_file->data_offset;
	if (dump_seek(device->dump_file, offset) < 0) {
		ret = SEEK_ERROR_IN_DUMP_FILE;
		goto out;
	}

	for (i = 0; i < device->header->max_mapnr; i++) {
		if (prog_flg)
			progress_bar(device->name, symbol, i,
					device->header->max_mapnr);

		if (!page_is_in_segs(i, device))
			continue;

		if (dump_is_compressed(device)) {
			if (page_is_dumpable(i, device))
				if ((ret = write_compress_pages(device, data))
						!= COMPLETE)
					goto out;
			continue;
		}

		if (page_is_dumpable(i, device)) {
			if (read(device->fd, data[0], block_size)
					< block_size) {
				ret = READ_ERROR_IN_DUMP_DEVICE;
				goto out;
			}
			if (check_zero_page(device, data[0]))
				goto skip;

			if (write(device->dump_file->fd, data[0], block_size)
					< block_size) {
				ret = WRITE_ERROR_IN_DUMP_FILE;
				goto out;
			}
		} else {
skip:
			if (lseek(device->dump_file->fd, block_size, SEEK_CUR)
					== failed) {
				ret = SEEK_ERROR_IN_DUMP_FILE;
				goto out;
			}
		}
	}

	/*
	 * Set the crash dump file size intentionally because the last block
	 * may not be actually written.
	 */
	if ((offset = lseek(device->dump_file->fd, 0, SEEK_CUR)) == failed) {
		ret = SEEK_ERROR_IN_DUMP_FILE;
		goto out;
	}

	if (ftruncate(device->dump_file->fd, offset) < 0) {
		ret = DUMP_FILE_SIZE_UNDETERMINED;
		goto out;
	}

	if (prog_flg)
		progress_bar_post(device->name, symbol, i,
				device->header->max_mapnr);

out:
	free(data[0]);
	free(data[1]);
	if (dump_is_compressed(device))
		device->header->status &= ~DUMP_HEADER_COMPRESSED;

	return ret;
}

static int blank_out_signature(DumpDevice *device)
{
	if (reopen_device_file(device) < 0)
		return FAILED;

	if (device_seek(device, (off_t)device->header->block_size) < 0)
		return FAILED;

	if (write(device->fd, ALREADY_SAVED_SIGNATURE, SAVE_SIG_LEN)
			< SAVE_SIG_LEN)
		return FAILED;

	return SUCCEEDED;
}

static inline void set_pagetype_flag(char *val)
{
	pagetype_flg = strtoull(val, NULL, 0);
}

static inline bool memory_page_completed(DumpDevice *device)
{
	return device->header->status == DUMP_HEADER_COMPLETED;
}

int main(int argc, char *argv[])
{
	DumpDevice *device;
	DumpFile *dump_file = NULL;
	int opt;
	bool verbose = false, del_flg = false, delete_only = false;
	char *dir;

	cmdname = CMD;

	while ((opt = getopt(argc, argv, "DfvVpP:X")) != -1) {
		switch (opt) {
		case 'D':
			del_flg = true;
			break;
		case 'f':
			frc_flg = true;
			break;
		case 'p':
			prog_flg = true;
			break;
		case 'v':
			verbose = true;
			break;
		case 'P':
			set_pagetype_flag(optarg);
			break;
		case 'V':
			version();
			break;
		case 'X':
			del_flg = true;
			delete_only = true;
			break;
		default:
			usage();
			break;
		}
	}

	if (optind + 1 != argc)
		usage();

	if ((device = create_new_device(argv[optind])) == NULL) {
		Err("unable to create a new device.\n");
		exit(1);
	}

	if (open_device_file(device) < 0) {
		Syserr("open %s", device->name);
		exit(1);
	}
	Info("%s\n", device->name);

	switch (read_device_header(device)) {
	case COMPLETE:
		break;
	case BLOCK_SIZE_INVALID:
		Err("block size is invalid.\n");
		exit(1);
	case SEEK_ERROR_IN_DUMP_HEADER:
		Syserr("seek the dump header");
		exit(1);
	case READ_ERROR_IN_DUMP_HEADER:
		Syserr("read the dump header");
		exit(1);
	case NO_PANIC_DUMP_HEADER:
		Err("%s does not have panic dump header.\n", device->name);
		exit(1);
	case BLOCK_SIZE_NOT_MATCH:
		Err("block size of %s does not match with system page size.\n",
								device->name);
		exit(1);
	case NUMBER_OF_CPUS_INVALID:
		Err("the number of CPU is invalid.\n");
		exit(1);
	case SEEK_ERROR_IN_DUMP_SUB_HEADER:
		Syserr("seek the dump sub header");
		exit(1);
	case READ_ERROR_IN_DUMP_SUB_HEADER:
		Syserr("read the dump sub header");
		exit(1);
	case SEEK_ERROR_IN_MEMORY_BITMAP:
		Syserr("seek the memory bitmap");
		exit(1);
	case READ_ERROR_IN_MEMORY_BITMAP:
		Syserr("read the memory bitmap");
		exit(1);
	default:
		Err("unknown error (reading the device header)\n");
		exit(1);
	}
	Info("  device header : read\n");

	if (del_flg && delete_only) {
		goto del_sig;
	}

	dir = make_dump_dir(device);
	if (dir == NULL) {
		Err("unable to make the dump directory\n");
		exit(1);
	}

	dump_file = open_dump_file(dir);
	if (dump_file == NULL) {
		Syserr("open the crash dump file");
		exit(1);
	}

	dump_file->device = device;
	device->dump_file = dump_file;
	Info("  dump file name: %s\n", dump_file->name);

	switch (check_fs(device)) {
	case -1:
		Err("unable to check %s.\n", dump_file->name);
		exit(1);
	case 0:
		break;
	case 1:
		Err("no space for dump image.\n");
		exit(1);
	default:
		Err("unknown error (checking file system)\n");
		exit(1);
	}
	Info("  file system   : available\n");

	build_segments_info(device);
	Info("  segment count : %d\n", device->num_segments);

	if (write_dump_header(device) < 0) {
		Err("unable to write the dump header.\n");
		exit(1);
	}
	Info("  dump header   : written\n");

	if (!prog_flg)
		Info("  memory dump   : ");

	switch (write_pages(device)) {
	case COMPLETE:
		break;
	case SEEK_ERROR_IN_DUMP_DEVICE:
		Syserr("seek the dump device");
		exit(1);
	case READ_ERROR_IN_DUMP_DEVICE:
		Syserr("read the dump device");
		exit(1);
	case WRITE_ERROR_IN_DUMP_FILE:
		Syserr("write to the crash dump file");
		exit(1);
	case SEEK_ERROR_IN_DUMP_FILE:
		Syserr("seek the crash dump file");
		exit(1);
	case DUMP_FILE_SIZE_UNDETERMINED:
		Syserr("set the crash dump file size");
		exit(1);
	case UNCOMPRESS_ERROR:
		Syserr("uncompress the crash dump file");
		exit(1);
	default:
		Err("unknown error (writing memory dump)\n");
		exit(1);
	}
	if (!prog_flg)
		Info("written\n");
	else
		Info("  memory dump   : written\n");
	Info("  written blocks: %u\n", device->header->written_blocks);

del_sig:
	if (del_flg) {
		if (is_swapon()) {
			Err("swap is running.\n");
			exit(1);
		}
		if (blank_out_signature(device) == FAILED) {
			Syserr("blank out dump signature");
			exit(1);
		}
		Info("  dump signature: blanked out\n");
		if (delete_only) {
			exit(0);
		}
	}

	if (!memory_page_completed(device)) {
		Err("dump partition has incomplete memory image.\n");
		exit(dump_is_short_area(device) ? 0 : 1);
	}

	if (rename_dump_file(dump_file) != 0) {
		Syserr("rename the crash dump file");
		exit(1);
	}
	Info("  dump file name: %s\n", dump_file->name);

	Info("  script to run : %s\n", DISKDUMP_SUCCESS);
	Info("  script status :\r");
	if (execute_script(DISKDUMP_SUCCESS, dump_file->dir) == 0) {
		Info("  script status : complete\n");
	} else {
		Info("  script status : incomplete or not found\n");
	}

	Info("saving panic dump completed.\n");
	exit(0);
}
